"""
    ExtendedCalendarCtrl.py

    A custom class using wx.adv.GenericCalendarCtrl with the ability to customise the calendar and the output format,
     based on some of the attributes of minidatepicker.py
    Works with wx.DateTime or python datetime values
    Uses wx.adv.GenericCalendarCtrl
    Uses wx.Locale to enable different languages for the calendar

    An attempt has been made to allow days marked with attributes denoting Holiday, Marked and Restricted
     to live with each other

    Holidays as far as the wx.CalendarCtrl is concerned seems to revolve round the weekends, for the most part,
     so for the purposes of this ExtendedCalendarCtrl, they are referred to as weekends and actual holidays,
     either defined by you or the python holidays module, are the actual holidays.
     The weekends are still marked by the attribute SetHoliday(True) because I cannot override it.

    Dates can be marked, restricted, defined as holidays and have notes.
     Marked Dates, Restricted dates and notes can be defined as a simple date or use more advanced rules for example
     the 3rd Friday of the month, every Tuesday of the month.or every 3rd Friday of every month
     Note: they can be year specific or every year

    Marked dates are marked with a Border, either Square or Oval, with optional Foreground/Background colours.
    Holidays are normally highlighted with a different Foreground/Background colour but only if you are Showing Holidays
    Restricted dates are marked using an Italic StrikeThrough font

    Defined Holidays can be year specific or fixed holidays i.e. every year on that Month/Day

    Public Holidays rely on the python 'holidays' module being available (pip install --upgrade holidays)
        public holidays are automatically entered into the Notes for you and the holiday module does make different
        different languages available, sometimes.
    You may add in more than one region's public holidays and they will be denoted by the country code
     and region code if appropriate.

    Notes are date specific or every year and in addition can follow the rules for Marked and Restricted dates, namely:
            All of a specified week day in a month;
            The specified occurrence of a weekday in a month e.g. 3rd Tuesday or last Friday of the month
            The specified occurrence of a weekday in every month e.g. 3rd Tuesday or last Friday of every month
            The last day of the month
            The last weekday of the month
        Dates with Notes are treated like Marked i.e. they can have a border, a text and background colour of their own

    Navigation:
     The Arrow keys will navigate the calendar
     The PageUp/PageDown keys will retreat and advance the month respectively, as will MouseScrollUp and MouseScrollDown
     The Home and End keys jump to the First and Last day of the month, respectively.
     A right click on the calendar, will display All the notes for the month.
     The F1 key is the keyboard equivalent of Right Click and will display All the notes for the month.
     Date ToolTips will be displayed as and when appropriate, depending of the position in the calendar and settings
     The F2 key opens a Year Calendar for reference.

    CalendarCtrl(parent, id=wx.ID_ANY, pos=wx.DefaultPosition, size=wx.DefaultSize,
                 style=0, name="ExtendedCalCtrl", date=0, formatter='', calendar_style=None, calendar_locale=None)

        @param parent:   Parent window. Must not be None.
        @param id:       identifier. A value of -1 indicates a default value.
        @param pos:      MiniDatePicker position. If the position (-1, -1) is specified
                         then a default position is chosen.
        @param size:     If the default size (-1, -1) is specified then a default size is calculated.
                         Size should be able to accomodate the specified formatter string + button
        @param style:    The style of the wx.Panel that the calendarctrl is set in.
        @param name:     Widget name.
        @param date:     Initial date (an invalid date = today)
        @param formatter A date formatting string in the form of a lambda function
                         The formatter will be called with a wx.DateTime thus we can use .Format()
                          the wxPython version of the standard ANSI C strftime
                         default lambda dt: dt.FormatISODate() = ISO 8601 format "YYYY-MM-DD".
                         or a lambda function with a format string e.g.:
                            lambda dt: (f'{dt.Format("%a %d-%m-%Y")}')
                         e.g.:
                            format = lambda dt: (f'{dt.Format("%a %d-%m-%Y")}')
                            format = lambda dt: (f'{dt.Format("%A %d %B %Y")}')
                         or
                            fmt = "%Y/%m/%d"
                            format = lambda dt: (dt.Format(fmt))
                            format = lambda dt: (dt.Format("%Y/%m/%d"))
                            format = lambda dt: (dt.FormatISODate())
                        for those who prefer strftime formatting:
                            format = (lambda dt: (f'{wx.wxdate2pydate(dt).strftime("%A %d-%B-%Y")}'))
        @param calendar_style
                        Sets the initial calendar style
                        Useful for styles not changeable after creation e.g. wx.adv.CAL_SEQUENTIAL_MONTH_SELECTION
        @param calendar_locale
                        Controls the Language used in the Calendar either None (default)
                        or a wx.Locale language e.g, wx.LANGUAGE_SPANISH or wx.LANGUAGE_GERMAN etc


    Events: EVT_DATE_CHANGED A date change occurred in the control
            existing wx.adv.GenericCalendarCtrl events are skipped, so you may bind to them if you wish

    Event Functions:
        GetValue()          Returns formatted date in the event as a string

        GetDate()           Returns wxDateTime date in the event, with all of its attendant functions

        GetNote()           Retrieve the note string of the date

        IsMarked()          True or False if the date is marked

        IsHoliday()         True or False if the date is a holiday

        IsPublicHoliday()   Retrieve the public holiday value of the date

        GetDateTime()       Returns python datetime of date in the event

        GetTimeStamp()      Returns seconds since Jan 1, 1970 UTC for current date


    CalendarCtrl Functions:

        GetValue()          Returns formatted date in the event as a string

        GetDate()           Returns wxDateTime date in the control

        GetDateTimeValue()  Returns python datetime of date in the control

        GetTimeStamp()      Returns seconds since Jan 1, 1970 UTC for selected date

        GetLocale()         Returns tuple of current language code and encoding

        GetNotes()          Returns the current Notes dictionary

        GetHolidays()       Returns the current Holidays dictionary

        GetRestrictedDate() Returns the current RestrictedDate dictionary

        GetMarkDates()      Returns the current MarkDates dictionary

        TriggerDateEvent(date=None)
                            Triggers a calendar event for the date specified or current date if no date supplied
                            The date passed can be a wx.Date or a datetime.datetime date

        SetValue(date)      Sets the date in the control
                            expects a wx.DateTime, a python datetime datetime or a timestamp
                            Any invalid date defaults to wx.DateTime.Today()

        SetFormatter(formatter) Date format in the form of a lambda
            default:    lambda dt: dt.FormatISODate()

        SetCalendarStyle(style)
            wx.adv.CAL_SUNDAY_FIRST: Show Sunday as the first day in the week
            wx.adv.CAL_MONDAY_FIRST: Show Monday as the first day in the week
            wx.adv.CAL_SHOW_HOLIDAYS: Highlight holidays in the calendar (only generic)
            wx.adv.CAL_NO_YEAR_CHANGE: Disable the year changing (deprecated, only generic)
            wx.adv.CAL_NO_MONTH_CHANGE: Disable the month (and, implicitly, the year) changing
            wx.adv.CAL_SHOW_SURROUNDING_WEEKS: Show the neighbouring weeks in the previous and next months
            wx.adv.CAL_SEQUENTIAL_MONTH_SELECTION: more compact, style for the month and year selection controls.
            wx.adv.CAL_SHOW_WEEK_NUMBERS

            Note: Some styles are Not changeable after the initial creation of the calendar

        SetCalendarHighlights(colFg, colBg)     Colours to mark the currently selected date

        SetCalendarHolidayColours(colFg, colBg) Colours to mark Holidays defaults None and Yellow

        SetWeekendColours(colFg, colBg)         Colours to mark weekends

        SetCalendarHeaders(colFg, colBg)        Colours for the calendar header area

        SetCalendarFg(colFg)    Set Calendar ForegroundColour

        SetCalendarBg(colBg)    Set Calendar BackgroundColour

        SetCalendarFont(font=None)  Set font of the calendar to a wx.Font
                Alter the font family, weight, size, etc

        SetCalendarMarkDates(markdates = {}) Mark dates with a Border
                A dictionary containing year and month tuple as the key and a list of days for the values to be marked
                e.g.
                    {
                     (2023, 7) : [2,5,7,11,30],
                     (2023, 8) : [7,12,13,20,27],
                     (2023, 9) : [1,27]
                    }

                Setting the year to zero in the key, will mark these dates every year
                Setting the month to zero, will mark these dates in every month

                Values of less than 0 indicate not a specific date but a day: -1 Monday, -2 Tuesday, ... -7 Sunday
                 allowing you to mark all Mondays and Fridays in the month of January e.g {(2023, 1) : [-1, -5]}
                 You may include a mixture of negative and positive numbers (days and specific dates)

                Negative values beyond that indicate the nth weekday, (the indexing is a bit confusing because it's off by 1
                 the first digit represents the day and the 2nd digit represents the occurrence i.e.

                -11, 1st Monday | -12, 2nd Monday | -13, 3rd Monday | -14, 4th Monday | -15, 5th or Last Monday
                -21, 1st Tuesday | -22, 2nd Tuesday | -23, 3rd Tuesday | -24, 4th Tuesday | -25, 5th or Last Tuesday
                ..............................................................................................................
                -71, 1st Sunday | -72, 2nd Sunday | -73, 3rd Sunday | -74, 4th Sunday | -75, 5th or Last Sunday

                This way all the individual days are grouped together.
                If the 5th occurrence of a weekday doesn't exist, the last occurrence of the weekday is substituted.

                -99 Stands for the last day of the month
                -98 is for the last weekday of the month

        SetCalendarMarkBorder(border=wx.adv.CAL_BORDER_SQUARE, bcolour=wx.NullColour, colFg=wx.NullColour, colBg=wx.NullColour)
                Defines the border type to mark dates wx.adv.CAL_BORDER_SQUARE (default)
                a border colour e.g. wx.NullColour (Default), wx.RED  or a hex value '#800080' etc
                a foreground colour
                and a background colour
                Valid border values are:
                    wx.adv.CAL_BORDER_NONE      - 0
                    wx.adv.CAL_BORDER_SQUARE    - 1
                    wx.adv.CAL_BORDER_ROUND     - 2

        SetCalendarHolidays(holidays = {})
                A dictionary containing year and month tuple as the key and a list of days for the values e.g.
                    {
                     (2023, 1) : [1,],
                     (2023, 7) : [1,30],
                     (2023, 8) : [7,15,27],
                     (2023, 12) : [25,26]
                    }

                Holidays can also be 'fixed' Holidays occurring every year on the same day by setting the year to zero in the key
                e.g.
                    {
                     (0, 1) : [1,],                            # January 1st is a Holiday every year
                     (2023, 7) : [1,30],
                     (2023, 8) : [7,15,27],
                     (0, 12) : [25,26]                         # Christmas Day and Boxing Day are Holidays every year
                    }

        SetCalendarNotes(notes = {})
                A dictionary containing a year, month, day tuple as the key and a string for the note e.g.
                    {
                     (2023, 1, 1) : "New Year's Day",
                     (2023, 12, 25) : "Christmas Day"
                    }

                Like Holidays, Notes can be assigned to a specific day every year, by setting the year to zero
                Setting the month to zero, will set the note on this date in every month
                    {
                     (0, 1, 1) : "New Year's Day",
                     (0, 12, 25) : "Christmas Day",
                     (2023, 0, -53) : "Pay day"                 # Every 3rd Friday in every month in 2023
                    }

                To compliment Marked Dates and Restricted Dates, notes can also be assigned a negative day following the
                 the same pattern as Marked Dates and Restricted Dates.
                Allowing you to match Notes with Marked Dates and Restricted Dates.

                    {
                     (0, 1, -11) : "The 1st Monday of January/the year",
                     (0, 1, -35) : "The last Wednesday of January",
                     (0, 2, -5)  : "Every Friday in February"
                    }

                If you set Public Holidays, they are enter automatically into the notes, marked with a leading asterix (*).

                Notes are displayed as a ToolTip, when the day is hovered over or Right clicked
                 or if the mouse is over the calendar or the Arrow keys are used to navigate the calendar to that day.

                A right click on the calendar, will display All the notes for the month, in a popup menu.
                Pressing F1 will do the same.

        SetCalendarNotesBorder(border=wx.adv.CAL_BORDER_SQUARE, bcolour=wx.GREEN, colFg=wx.NullColour, colBg=wx.NullColour))
                Defines the border type to dates with notes wx.adv.CAL_BORDER_SQUARE (default)
                a border colour e.g. wx.GREEN (Default), wx.RED  or a hex value '#800080' etc
                a foreground colour
                and a background colour
                Valid border values are:
                    wx.adv.CAL_BORDER_NONE      - 0
                    wx.adv.CAL_BORDER_SQUARE    - 1
                    wx.adv.CAL_BORDER_ROUND     - 2

        SetCalendarRestrictDates(rdates = {})
                A dictionary containing a year and month tuple as the key and a list of days, for the days that
                 are Not selectable within that year/month
                 e.g.
                    {
                     (2023, 1) : [1,15],
                     (2023, 3) : [1,15],
                     (2023, 5) : [1,15],
                     (2023, 7) : [1,15,23],
                     (2023, 9) : [1,15],
                     (2023, 11) : [1,15]
                    }

                All dates in the 'restricted' dictionary use an Italic StruckThrough font and cannot be selected

                See SetCalendarMarkDates for the ability to use negative values to calculate dictionary values to restrict
                more complicated entries like All Mondays or the 2nd and 4th Tuesday for example, by using negative values.

        SetCalendarDateRange(lowerdate=wx.DefaultDateTime, upperdate=wx.DefaultDateTime)
                Either 2 wx.DateTime values to restrict the selectable dates
                or just a lower date or just an upper date
                (The oddity of wx.DateTime months from 0 is catered for)
                Returns False if the dates are not wx.DateTime objects
                wx.DefaultDateTime equals no date selected.

                Dates outside of the range will display an "Out of Range" ToolTip, with the defined range.

        SetCalendarOnlyWeekDays(boolean) Default False
                If set only weekdays are selectable. weekends and holidays use an Italic StruckThrough font and cannot be selected
                Holidays are treated as Not a weekday i.e. no work

        AddPublicHolidays(country='', subdiv='', language='', replace=False) Default blank, blank, blank, False
                Only available if the python 'holidays' module was successfully imported
                Currently supports 134 country codes using country ISO 3166-1 alpha-2 codes and the optional subdivision
                 (state, region etc) using ISO 3166-2 codes.
                Language must be an ISO 639-1 (2-letter) language code, except the ubiquitous 'en_US'.
                 If the language translation is not supported the original holiday names are returned.
                The replace parameter if set to True, will create a new set of public holidays, rather than adding
                 to the existing set. Aimed at replacing public holidays, based on circumstances.

                For details: https://python-holidays.readthedocs.io/en/latest/
                 (or the file 'python-holidays — holidays documentation.html' supplied with this program)
                e.g.
                    country='ES'                    Spain
                    country='ES' and subdiv='AN'    Spain, Andalucia
                    country='GB' and subdiv='ENG'   Great Britain, England
                    country='US' and subdiv='SC'    USA, South Carolina

                function returns True if successful, for an existing country and subdivision (if supplied)
                 or False if there was an error

                This function can be called multiple times, once for each country or region in a country
                 that you wish marked on the calendar, unless the replace parameter is set to True.
                The first call sets the primary holiday region.
                May be useful if you are operating in more than one geographical area, with differing holidays,
                 or you may wish to display the holidays once in the original language and again in English.
                 e.g.
                    AddPublicHolidays(country="DE", subdiv="ST", language=())         # Germany region Lower Saxony
                    AddPublicHolidays(country="DE", subdiv="ST", language=('en_US'))  # In English

Author:     J Healey
Created:    16/09/2023
Copyright:  J Healey - 2023
License:    GPL 3 or any later version
Email:      <rolfofsaxony@gmx.com>
Version     1.0
"""

import wx
import wx.adv
import datetime
import calendar
try:
    import holidays
    holidays_available = True
except ModuleNotFoundError:
    holidays_available = False

from wx.lib.embeddedimage import PyEmbeddedImage

cal_event = PyEmbeddedImage(
    b'iVBORw0KGgoAAAANSUhEUgAAABkAAAAZCAYAAADE6YVjAAAABmJLR0QARQD3AAo5DuANAAAC'
    b'AElEQVRIx83VTYiNYRQH8N975465zYgISU2yMJORj2gWdmKprNgrSx8lpeyGhZUsfGzIShKT'
    b'hRILCwspoix8ra18JDJMkxvva3Pu9Hh7773cW3Lqqaf3Oe//f87/Oec8/AdWQ72Lz0CsnmwA'
    b'h3EDE218luISzmC4F5KFeIccJ5BV+OxGE3OYbAdUr5BncaxZLAjwIazASOLbRCPOa+GzGjP4'
    b'EsHNg7ZsEMfwEs+wuSTdKTxP1q0AbtkIHuJV4AxWZbILU/FjsxSAyCrVvVGhzDCWBM7rCGQe'
    b'qIYdpcgy/Ix9XkFaSyQpYt+6t6HAq5XlapTkGcd1vMADrCuRrMR7PMb9kKdRhVe++O9JdEdw'
    b'GvewB2NRRZJoj+NsSDkVGc2Vs86S1M/hKR4l53msTg35I3DShtyGrTiEvPzzm7iwfm0VtlSV'
    b'cC+2BgewrJNTvQ+CtVEYE/iA6U4DsNuAHKkYgGMx0zYG+N1uIJ1se0yAo0kHjwfBBlwLub71'
    b'I1cTi3AyArqNq1ifEMz8zbtxIbq0/H0vPkcPvY2SvRLk7WwnzreU6pZJjpuxv4jlvWSQkhRt'
    b'3owW0Sw2RYRfu+BmgfcbSYFPGO2Q0Z2oouIPgh8NvKJMMh2j5WN0fdFD72QxSPfhYAsjKzlM'
    b'Yn9M2KwHkiKe7Mt4UkWSjvl+xk2evEP/zn4BOsdrpbkFvrcAAAAASUVORK5CYII=')

cal_noevent = PyEmbeddedImage(
    b'iVBORw0KGgoAAAANSUhEUgAAABkAAAAZCAYAAADE6YVjAAAABmJLR0QA/wD/AP+gvaeTAAAD'
    b'PklEQVRIx82WT2hVRxTGfzNzc/OYpKRGK1ppotDYaBcqGoIbg3HTUhCEEkpdhUIXFQttoail'
    b'JS4UCtqFZqWCcdEGU5W24B9w0VIb6R/EiPoU0SIJpUqL1URv4subGRfvvMfz9j3D040HBi7M'
    b'me8755s551x4DkwD0Qw+RtZTmQE+BIaApVV8moH9wFeAfRqSRuAW4IHtgKrgsx7IAZNARzWg'
    b'qII8TbIeALGA1wNzgYYy3xyQkX0tPq3AOHBPgiuBFq0O+BS4DJwHlqek2wlcLFvfC3DRGoBh'
    b'ICs4dZUyeQvok4O5VABIVuW6ZyooY4FZgnNFAikBaaA7FZkCnHz7CqS6TJIg38V7qxc8nZYr'
    b'k5LnNeAwcAk4AyxJkcwDbgO/AT+KPJlKeCW5upQyLZCbAP9dCAH4CNgFnAbeBhbLK6Is2q3A'
    b'HpGyTzKaTGddTE9/ovXBjVHk52idPTE9feZr5yYO1Nd3nsznz//i/fSREFyVF5oXnPKCXA2s'
    b'BDYDvkTSAv07omj43TieD3QB/8leFrhmkuRYDTW2DtggxexLaY1C2J3P3xJZOiT1NqAHWOSs'
    b'XeWsbQRw1r4qxxYBm4A5M/Wmko0UgAeAXuAF2T8HvAJsBNqctYuBzDtKtQNHgd3A2ieR/K/5'
    b'mSQ566xtBhZ6uK3gC+APVchqBLhy0/sVmYLe7cAgcLImErG7wGDf1NSaNq1fXmHMgteNua+g'
    b'Iwlhg4HuTq0b/vF+6HgIm4D7NZOYJPFCdOeNEOp+d25dlzED6+O4M+tc6zQ0ro2i6z1af/Oi'
    b'Ui3ATcCbJElqyaRoZ0+F8D6wL3Gu9++HDycued/0ZSZzxyr1Q1y4+DelhZxz1l4Fxk2KayYS'
    b'L5fLQAj7ljrXNBuGFByO4S/guHSFVmAhMFb0r0YSqsyMItGDLCwD+puUmpQq/0Be4OdC0AmM'
    b'LYPxCwW8xypeyWD6U55wNVOUHXbWRkJ2GvhJ5tDquyFEH09NnToUwhYgRGVZfAvsBf4VjcOM'
    b'8zlJ6FZKb4uiLXO1bm43Ztd4CKODudzse4WnHUjJo6TS35MOq2qZ1UtAfRZFL13w/sblEPpP'
    b'hPDrkwI10rZrXj/HccOz/Lk8kz0C/Mb+m+9IOV0AAAAASUVORK5CYII=')

cal_fiesta = PyEmbeddedImage(
    b'iVBORw0KGgoAAAANSUhEUgAAABkAAAAZCAYAAADE6YVjAAAABmJLR0QARQD3AAo5DuANAAAG'
    b'WElEQVRIx5WWyY9cVxXGf/e9+2p6XVU9lnvuTrvbTrAdx9hRwE5sJ0KKUIwiocgLvEXiP0As'
    b'Yc2/wIIFwwJWCARCIFAUBRO17aSJ3W73ZPdQXfPwanjzvSyq7dhGQnCkq6O7Od8953463yd4'
    b'MSRgXLu+yE9++kNrsz2retWy+vu9Le1Uq6z+6Y84pTKgQWv+Sygger4ogAFcBr6fHkpP940Z'
    b'8ee/bqSVvyYPan0/5fVo9XVw5utn2L5vUTkogVag1CC/GBooAj8DPgWUABCCty0pfqu0KCwv'
    b'F8T1b79Hr90Ae4JUWrL5+T1MSzIzPUpPnuDOP+7jVEo0m110FEIcgIpfBqoAHwGfmICUpvjx'
    b'0mL28vzX3hA3PnyHbten1XEolxrs7DVgeI6OzmBYSS5cPMnk+Djdnk+95aMMC4zjgehnQAKw'
    b'ARP4vTmUMaXS+gd1R5/8zs33eW1GEQVdCgvLNOtt2l3NuTMLxIFPJCRxrJB+A88ax/A7NF2B'
    b'FhIMc/B+FR83ggCqwK/lW5dmqdU7Ij3zBrYlWH3gUJgq8NmdPZLZSeKDXT65vYtlauzJaaL9'
    b'DitLk1xeSPBPDLZqXww6MczjukDYf0oMASA//nQPhcH7Z4dQQZdq3yLd6uA5Dg8362QTGiub'
    b'RbgdbDPC8xX3d1vUqppqqcLs8Aj77TYIcfwZejC20PuKsmGkMSwTQ4XEymA+1+LeRoyYWOLa'
    b'Up/q4yMOQs3UK9PUKk3mlqeI3B5SCKYWpsgYVV5fniPwXar1Co7T4qCoiLQijoKvKHx2ZYiD'
    b'ZsCJScHMiRzdR20qTyq8+8EKlf0mEymB48YYUlDvakasFIbuUzrw+NaFd/ADRRh2SSbyjJ4a'
    b'wnMbrN3/nK3tDcIwHIBsPXaQjXVydpqNUprz82myGYPCmCCblRSGQ3biNNnZLHUnwjnaY3px'
    b'ghvfvMLefpVI+ewVy4zlBY/2drlx9Roj+SzZTIrP7q5iAmaojFuhlkuGZfGNK+d582yenaM+'
    b'20UPbSXJTozx8b9qeJ0Oudjl/LkCr5+7RCqcwQ8DolBTqhXJpYZJWAZRHGIKWJyf3n37ysVf'
    b'mghhYshbWqaX+gFsbpXZKUe0OjFWOoPf6yPSOZJunSE7w8lpye27HZYmX+Wo9IS247O++5Ao'
    b'VDScEmEY43ktLFMykhs5unnzo18ZA5YJEIL5MUngB+xvl9CxJm3nOLsyScprs7g4g+p0eLQR'
    b'4PcCaodVXFfj9DqMDWdJygxCmCQsSSaVpdfvoSItx+180kRgYhi3kMmlhmcRCIt+bFJtejhN'
    b'F6kypJWFKvvMjY5RLLdxA8XoSJqMzHB7fY1ziytIkSHGx+l6DKUTqNjHdbsbjXbp5wZaD5ac'
    b'jo+zZlgKLo7YnEorriYs5uodZsOQRMvl+olxXrOT3H3wCDds8b2rHzI3tQCGwiDF8vwJEgmB'
    b'bU0wVZiWnU5HD5aOUhBHoCKkjjmZTXFhdoJqrU3YctgutWgaBivjeZSAdxcm8Q9qNLsuSWuP'
    b'jYf7TI+OYyVyHJWPGLNzIDy6XhD+5nd/OF71WkEcQhwyRMx7r75CVkFiJIcWMDtTwGt36bs+'
    b'o6kEvusxk7b4xeqXzNgJLp07RUqaZOQQYW6YZvOIau0J99fXCcPomZ5AHCAiH1soOr0uPT+i'
    b'kM9zWGnSS1g4nT57WiClpCw0bhRzNp9ip+Pxt9U1lgpZVByzufOEQj7Nl+sPnq3/r0BUTEIF'
    b'vDk1xs7+IVGkKXoBAoOdSoOkJclJyRe1FlN2kulshqShSRFx6PQo1msQeuigR6XsvKAvxvNK'
    b'4/t9HlXKnBzJIlXE6bEcLbdP3oRat0dGGqSlYMZO4gU+B402e45DFHro0EOHffB7EPkvSKXx'
    b'nJKB1hTrVSw7ybgFW4eHzA8lGZKa2YyFqUJsHfK4eMRus8m246ADF/w+BD3wuhC6z+u/fjou'
    b'dazJGhCoGM/3yY+PIKKYWEdYUcBKPovf9/F9n640WTs8RAceROHg5aE7yC8CFAFlHoPUgA8A'
    b'24siUXMclCXJmiYoxWTeRhiChcI4ndBn7bDIXrMJQX9wwv6Anf+p8T8CHouX3QowDQg7keCt'
    b'06c5NTeHnUjgRpGxX62pO5tbFB3n/3crL/uup5d8KsV3r13jL6urnFlcpOo41FptntRqx2PR'
    b'/5Pv+jdB5lKV7c0slAAAAABJRU5ErkJggg==')

__version__ = 1.0


exCalEVT = wx.NewEventType()
EVT_DATE_CHANGED = wx.PyEventBinder(exCalEVT, 1)

# day ranges for negative day codes see GenerateDates and GenerateNotes
mondays     = range(-11, -16, -1)
tuesdays    = range(-21, -26, -1)
wednesdays  = range(-31, -36, -1)
thursdays   = range(-41, -46, -1)
fridays     = range(-51, -56, -1)
saturdays   = range(-61, -66, -1)
sundays     = range(-71, -76, -1)


class exCalEvent(wx.PyCommandEvent):
    def __init__(self, eventType, eventId=1, date=None, note=None, marked=None, holiday=None, public_holiday=None, value=''):
        """
        Default class constructor.

        :param `eventType`: the event type;
        :param `eventId`: the event identifier.
        """
        wx.PyCommandEvent.__init__(self, eventType, eventId)
        self._eventType = eventType
        self.date = date
        self.note = note
        self.marked = marked
        self.holiday = holiday
        self.public_holiday = public_holiday
        self.value = value

    def GetDate(self):
        """
        Retrieve the date value of the control at the time
        this event was generated, Returning a wx.DateTime object"""
        return self.date

    def GetNote(self):
        """
        Retrieve the note value of the control at the time
        this event was generated, Returning a string"""
        return self.note

    def IsMarked(self):
        """
        Retrieve the marked value of the control at the time
        this event was generated, Returning a string"""
        return self.marked

    def IsHoliday(self):
        """
        Retrieve the holiday value of the control at the time
        this event was generated, Returning a string"""
        return self.holiday

    def IsPublicHoliday(self):
        """
        Retrieve the public holiday value of the control at the time
        this event was generated, Returning a string"""
        return self.public_holiday

    def GetValue(self):
        """
        Retrieve the formatted date value of the control at the time
        this event was generated, Returning a string"""
        return self.value.title()

    def GetDateTime(self):
        """
        Retrieve the date value of the control at the time
        this event was generated, Returning a python datetime object"""
        return wx.wxdate2pydate(self.date)

    def GetTimeStamp(self):
        """
        Retrieve the date value represented as seconds since Jan 1, 1970 UTC.
        Returning a integer
        """
        return int(self.date.GetValue()/1000)



class CalendarCtrl(wx.Panel):
    def __init__(self, parent, id=wx.ID_ANY, pos=wx.DefaultPosition, size=wx.DefaultSize,
                 style=0, name="ExtendedCalCtrl", date=0, formatter='', calendar_style=None, calendar_locale=0):
        wx.Panel.__init__(self, parent, id, pos=pos, size=size, style=style, name=name)
        self.parent = parent
        self._date = date
        if formatter:
            format = formatter
        else:
            format = lambda dt: dt.FormatISODate()
        self._formatter = format
        font = wx.SystemSettings.GetFont(wx.SYS_SYSTEM_FONT)
        self.SetWindowStyle(wx.BORDER_NONE)
        self.locale =  wx.Locale(calendar_locale, wx.LOCALE_LOAD_DEFAULT)
        self._style = style
        if calendar_style:
            self._cal_style = calendar_style
        else:
            self._cal_style = wx.adv.CAL_MONDAY_FIRST|wx.adv.CAL_SHOW_HOLIDAYS
        self.ShowHolidays = True
        self._cal_headercolours = (wx.NullColour, wx.NullColour)
        self._cal_highlightcolours = (wx.NullColour, wx.NullColour)
        self._cal_holidaycolours = (wx.NullColour, wx.YELLOW)
        self._cal_Bg = wx.NullColour
        self._cal_Fg = wx.NullColour
        self._cal_Font = font
        self._cal_MarkDates = {}
        self._cal_MarkBorder = (wx.adv.CAL_BORDER_SQUARE, wx.NullColour, wx.NullColour, wx.NullColour)
        self._cal_MarkNotes = (wx.adv.CAL_BORDER_SQUARE, wx.GREEN, wx.NullColour, wx.NullColour)
        self._cal_Holidays = {}
        self._cal_RestrictDates = {}
        self._cal_daterange = (wx.DefaultDateTime, wx.DefaultDateTime)
        self._cal_Notes = {}
        self._cal_SetOnlyWeekDays = False
        self._cal_PublicHolidays = False
        self._cal_AddPublicHolidays = []

        self.calendar = wx.adv.GenericCalendarCtrl(self, style=self._cal_style)
        date = wx.DateTime.Today()
        date = date.ResetTime()
        self.calendar.SetDate(date)
        self.Holidays = {}
        self.PublicHolidays = {}
        self.RestrictedDates = {}
        self.MarkDates = {}
        self.Notes = {}

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.calendar, 0, wx.ALL|wx.EXPAND, 5)
        self.SetSizer(sizer)

        self.calendar.Bind(wx.adv.EVT_CALENDAR_MONTH, self.OnChange)
        self.calendar.Bind(wx.adv.EVT_CALENDAR_YEAR, self.OnChange)
        self.calendar.Bind(wx.adv.EVT_CALENDAR, self.OnChosen)
        self.calendar.Bind(wx.adv.EVT_CALENDAR_SEL_CHANGED, self.OnToolTip)
        self.calendar.Bind(wx.EVT_MOTION, self.OnToolTip)
        self.calendar.Bind(wx.EVT_RIGHT_DOWN, self.OnRightClick)
        self.calendar.Bind(wx.EVT_KEY_DOWN, self.OnKey)
        self.Show()

    def SetFormatter(self, formatter):
        '''formatter will be called with a wx.DateTime'''
        self._formatter = formatter
        self.OnDate(self._date)

    def SetCalendarStyle(self, style=0):
        self._cal_style = style
        self.calendar.SetWindowStyleFlag(style)
        if style & wx.adv.CAL_SHOW_HOLIDAYS or style == wx.adv.CAL_SHOW_HOLIDAYS:
            self.ShowHolidays = True
        else:
            self.ShowHolidays = False

    def SetCalendarHeaders(self, colFg=wx.NullColour, colBg=wx.NullColour):
        self._cal_headercolours = colFg, colBg
        self.calendar.SetHeaderColours(self._cal_headercolours[0],self._cal_headercolours[1])

    def SetCalendarHighlights(self, colFg=wx.NullColour, colBg=wx.NullColour):
        self._cal_highlightcolours = colFg, colBg
        self.calendar.SetHighlightColours(self._cal_highlightcolours[0],self._cal_highlightcolours[1])

    def SetCalendarHolidayColours(self, colFg=wx.NullColour, colBg=wx.NullColour):
        self._cal_holidaycolours = colFg, colBg

    def SetWeekendColours(self, colFg=wx.NullColour, colBg=wx.NullColour):
        self._cal_weekendcolours = colFg, colBg
        self.calendar.SetHolidayColours(colFg, colBg)

    def SetCalendarFg(self, colFg=wx.NullColour):
        self._cal_Fg = colFg
        self.calendar.SetForegroundColour(self._cal_Fg)

    def SetCalendarBg(self, colBg=wx.NullColour):
        self._cal_Bg = colBg
        self.SetBackgroundColour(self._cal_Bg)

    def SetCalendarFont(self, font=None):
        self._cal_Font = font
        self.calendar.SetFont(self._cal_Font)
        bz = self.calendar.GetBestSize()
        self.calendar.SetSize(bz)
        self.Refresh()

    def SetCalendarMarkDates(self, markdates = {}):
        self._cal_MarkDates = markdates
        self.SetMarkDates(self._cal_MarkDates)

    def SetCalendarMarkBorder(self, border=wx.adv.CAL_BORDER_SQUARE, bcolour=wx.NullColour,\
                              colFg=wx.NullColour, colBg=wx.NullColour):
        self._cal_MarkBorder = (border, bcolour, colFg, colBg)

    def SetCalendarMarkNotes(self, border=wx.adv.CAL_BORDER_SQUARE, bcolour=wx.GREEN,\
                             colFg=wx.NullColour, colBg=wx.NullColour):
        self._cal_MarkNotes = (border, bcolour, colFg, colBg)

    def SetCalendarHolidays(self, holidays = {}):
        self._cal_Holidays = holidays
        self.SetHolidays(self._cal_Holidays)

    def AddPublicHolidays(self, country='', subdiv='', language='', replace=False):
        if not holidays_available:
            return False
        country = country.upper()
        subdiv = subdiv.upper()
        try:
            supported = holidays.country_holidays(country=country).subdivisions
        except Exception as e:
            return False
        if subdiv:
            if subdiv not in supported:
                return False
        if replace: # Do not add to existing public holidays, create a new set.
            self._cal_PublicHolidays = False
        if not self._cal_PublicHolidays:
            self._cal_PublicHolidays = (country, subdiv, language)
        else:
            self._cal_AddPublicHolidays.append([country, subdiv, language])
        self.SetPublicHolidays(self._cal_PublicHolidays)
        return True

    def SetCalendarRestrictDates(self, rdates = {}):
        self._cal_RestrictDates = rdates
        self.SetRestrictDates(self._cal_RestrictDates)

    def SetCalendarDateRange(self, lowerdate=wx.DefaultDateTime, upperdate=wx.DefaultDateTime):
        if not isinstance(lowerdate, wx.DateTime):
            return False
        if not isinstance(upperdate, wx.DateTime):
            return False
        self._cal_daterange = (lowerdate, upperdate)
        self.SetDateRange(self._cal_daterange[0], self._cal_daterange[1])

    def SetCalendarNotes(self, notes = {}):
        self._cal_Notes = notes
        self.SetNotes(self._cal_Notes)

    def SetCalendarOnlyWeekDays(self, wds = False):
        self._cal_SetOnlyWeekDays = wds
        if wds:
            self._cal_style = self._cal_style | wx.adv.CAL_SHOW_HOLIDAYS
            self.ShowHolidays = True

    def TriggerDateEvent(self, date=None):
        if not date:
            date = self.calendar.GetDate()
        if isinstance(date, datetime.datetime):
            date = wx.DateTime.FromDMY(date.day, date.month-1, date.year)
        self.OnDate(date)

    def OnDate(self, date):
        self._date = date
        d = (date.year, date.month + 1, date.day)
        ym = (date.year, date.month + 1)
        note = self.Notes.get(d, '')
        marked = self.MarkDates.get(ym, [])
        if date.day in marked:
            marked = True
        else:
            marked = False
        hol = self.Holidays.get(ym, [])
        if hol:
            pass
        else:
            hol = self.Holidays.get((0, date.month + 1), [])
        if date.day in hol:
            hol = True
        else:
            hol = False
        ohol = self.PublicHolidays.get(ym, [])
        if date.day in ohol:
            ohol = True
        else:
            ohol = False
        event = exCalEvent(exCalEVT, self.GetId(), date=date, note=note, marked=marked,\
                         holiday=hol, public_holiday=ohol, value=self._formatter(date))
        event.SetEventObject(self)
        self.GetEventHandler().ProcessEvent(event)

    def OnKey(self, event):
        keycode = event.GetKeyCode()
        if keycode == wx.WXK_F1:
            self.OnRightClick(None)
        if keycode == wx.WXK_F2:
            d = self.calendar.GetDate()
            if self._cal_style & wx.adv.CAL_SUNDAY_FIRST:
                fwd = 6
            else:
                fwd = 0
            ycal = calendar.TextCalendar(firstweekday=fwd).formatyear(d.year, c=3)
            win = YearPopup(self, txt=ycal)
            win.Show()
        event.Skip()

    def GetValue(self):
        #return self.calendar.GetValue()
        return self._formatter(self.calendar.GetDate())

    def GetDate(self):
        return self.calendar.GetDate()

    def GetDateTimeValue(self):
        """
        Return a python datetime object"""
        return wx.wxdate2pydate(self.calendar.GetDate())

    def GetTimeStamp(self):
        """
        Retrieve the date value represented as seconds since Jan 1, 1970 UTC.
        Returning a integer
        """
        return int(self.calendar.GetDate().GetTicks())

    def GetLocale(self):
        return self.locale.GetLocale()

    def GetNotes(self):
        return self.Notes

    def GetHolidays(self):
        return self.Holidays

    def GetMarkDates(self):
        return self.MarkDates

    def GetRestrictedDates(self):
        return self.RestrictedDates

    def SetValue(self, date):
        if isinstance(date, wx.DateTime):
            pass
        elif isinstance(date, datetime.date):
            date = wx.pydate2wxdate(date)
        elif isinstance(date, int) and date > 0:
            date = wx.DateTime.FromTimeT(date)
        elif isinstance(date, float) and date > 0:
            date = wx.DateTime.FromTimeT(int(date))
        else:  # Invalid date value default to today's date
            date = wx.DateTime.Today()
        self._date = date.ResetTime()
        self.SetFormatter(self._formatter)

    def OnChosen(self, event=None):
        ''' Test chosen date for inclusion in restricted dates if set
            Test if set to only allow weekdays, test if it is a weekday or a holiday, which is treated as not a weekday
        '''
        d = self.calendar.GetDate()
        event.Skip()
        if self.RestrictedDates:
            test = (d.year, d.month+1)
            days = self.RestrictedDates.get(test, ())
            if not days or d.day not in days:
                pass
            else:
                return

        if self._cal_SetOnlyWeekDays and not d.IsWorkDay(): # Weekend
            return
        if self._cal_SetOnlyWeekDays: # Holiday
            attr = self.calendar.GetAttr(d.day)
            if attr.IsHoliday():
                return

        self.calendar.GetDate()
        self.OnDate(self.calendar.GetDate())

    def OnChange(self, event):
        # If the year changed, recalculate the dictionaries for Marked, Restricted, Public Holidays and Note dates
        if event.GetEventType() == wx.adv.EVT_CALENDAR_YEAR.typeId:
            self.MarkDates = self.GenerateDates(self._cal_MarkDates)
            self.RestrictedDates = self.GenerateDates(self._cal_RestrictDates)
            self.SetPublicHolidays(self._cal_PublicHolidays)
            self.Notes = self.GenerateNotes(self._cal_Notes)
        date = event.GetDate()
        self.OnMonthChange()
        # Permit calling programs to see Month and Year Events
        event.Skip()

    def SetMarkDates(self, markdates):
        self.MarkDates = self.GenerateDates(markdates)
        self.OnMonthChange()

    def OnMonthChange(self):
        font = self.calendar.GetFont()
        font.SetStrikethrough(True)
        font.MakeItalic()
        date = self.calendar.GetDate()
        days_in_month = date.GetLastMonthDay().day
        mark_days = self.MarkDates.get((date.year, date.month+1), [])    # get dict values or an empty list if none
        h_days = self.Holidays.get((date.year, date.month+1), [])
        fixed_h_days = self.Holidays.get((0, date.month+1), [])
        r_days = self.RestrictedDates.get((date.year, date.month+1), [])
        oh_days = self.PublicHolidays.get((date.year, date.month+1), [])

        if isinstance(mark_days, int): # Allow for people forgetting it must be a tuple, when entering a single day
            mark_days = tuple((mark_days,))
        if isinstance(h_days, int):
            h_days = tuple((h_days,))
        if isinstance(fixed_h_days, int):
            fixed_h_days = tuple((fixed_h_days,))
        if isinstance(r_days, int):
            r_days = tuple((r_days,))

        mark_border, mb_clr, mb_fg, mb_bg = self._cal_MarkBorder    #border type, border colour, background colour, text colour
        mark_notes, mn_clr, mn_fg, mn_bg = self._cal_MarkNotes

        for d in range(1, days_in_month+1):
            note = self.Notes.get((date.year, date.month +1, d), '')            # get notes for the day
            attr = self.calendar.GetAttr(d)
            highlight_attr = wx.adv.CalendarDateAttr()
            if d in mark_days and mark_border:                                  # Marked Day & borders on
                highlight_attr.SetBorder(mark_border)
                highlight_attr.SetBorderColour(mb_clr)
                highlight_attr.SetBackgroundColour(mb_bg)
                highlight_attr.SetTextColour(mb_fg)
            if note and mark_notes:                                             # Note for the day & borders on
                highlight_attr.SetBorder(mark_notes)
                highlight_attr.SetBorderColour(mn_clr)
                highlight_attr.SetBackgroundColour(mn_bg)
                highlight_attr.SetTextColour(mn_fg)
            if d in h_days or d in fixed_h_days or d in oh_days:            # Holiday/fixed holiday/public holiday
                highlight_attr.SetTextColour(self._cal_holidaycolours[0])
                highlight_attr.SetBackgroundColour(self._cal_holidaycolours[1])
            if self.ShowHolidays:
                if not wx.DateTime(d, date.month, date.year).IsWorkDay():       # Weekend
                    highlight_attr.SetHoliday(True)
            if d in r_days:                                                     # Resticted Day (override holiday)
                highlight_attr.SetFont(font)
            if highlight_attr.IsHoliday():
                if self._cal_SetOnlyWeekDays:
                    highlight_attr.SetFont(font)
            if highlight_attr is not None:
                self.calendar.SetAttr(d, highlight_attr)
            else:
                self.calendar.ResetAttr(d)

        self.calendar.Refresh()

    def SetHolidays(self, holidays):
        self.Holidays = holidays
        self.OnMonthChange()

    def SetPublicHolidays(self, holiday_codes):
        self.PublicHolidays = {}
        if not holiday_codes:                                                   # holiday codes not set
            return
        country, subdiv, language = holiday_codes
        self.country_name = country
        for c in holidays.registry.COUNTRIES.values():
            if country in c:
                self.country_name = c[0]
                break
        d = self.calendar.GetDate()
        for k, v in holidays.country_holidays(country=country, subdiv=subdiv, years=d.year, language=language).items():
            existing = self.PublicHolidays.get((k.year, k.month), [])
            if k.day not in existing:
                self.PublicHolidays[(k.year, k.month)] = existing + [k.day]

        for item in self._cal_AddPublicHolidays:
            country, subdiv, language = item
            for k, v in holidays.country_holidays(country=country, subdiv=subdiv, years=d.year, language=language).items():
                existing = self.PublicHolidays.get((k.year, k.month), [])
                if k.day not in existing:
                    self.PublicHolidays[(k.year, k.month)] = existing + [k.day]
        self.SetNotes(self._cal_Notes)                                 # Update Notes with public holidays
        self.OnMonthChange()

    def SetDateRange(self, lowerdate=wx.DefaultDateTime, upperdate=wx.DefaultDateTime):
        if lowerdate.IsValid() or upperdate.IsValid():
            if lowerdate.IsValid():
                lowerdate = wx.DateTime(lowerdate.day, lowerdate.month-1, lowerdate.year)
            if upperdate.IsValid():
                upperdate =  wx.DateTime(upperdate.day, upperdate.month-1, upperdate.year)
            self.calendar.SetDateRange(lowerdate, upperdate)

    def SetNotes(self, notes):
        self.Notes = self.GenerateNotes(notes)
        self.OnMonthChange()

    def SetRestrictDates(self, rdates):
        self.RestrictedDates = self.GenerateDates(rdates)
        self.OnMonthChange()

    def restricted_date_range(self, start, end):
        '''
            Generate dates between a start and end date
        '''
        for i in range((end - start).days + 1):
            yield start + datetime.timedelta(days = i)

    def day_in_range(self, start, end, day):
        '''
            Test if date is the required day of the week
        '''
        for d in self.restricted_date_range(start, end):
            if d.isoweekday() == day:
                yield d

    def GenerateDates(self, date_dict):
        ''' Generated on start and when the year changes (Marked and Restricted dictionaries)
            This routine generates a new dictionary of complete dates, from the one passed in
             and returns the generated dictionary.
            This is because the original passed in dictionary may include date codes e.g. -99 for the last day of a month
             or -1 all Mondays or -23 the 3rd Tuesday, which need to be calculated for the given month in the given year.
            An added complication is that the year may be set to zero, denoting all years, so if the calendar year is
             changed, this routine is run again, to ensure that the dates are relevant to the current year.
        '''
        generated_dict = {}

        for year, month in date_dict:
            gen_year = year
            if gen_year == 0:               # Zero entry = All years, so generate dates for the currently selected year
                d = self.calendar.GetDate()
                gen_year = d.year
            if month <= 0:                  # Zero entry = All months, generate dates for every month in the year
                month_start = 1
                month_end = 13
            else:
                month_start = month
                month_end = month + 1
            for month in range(month_start, month_end):
                day_map = calendar.monthcalendar(gen_year, month)
                for neg in list(date_dict.get((year, month))):
                    if neg >= 0:
                        existing = generated_dict.get((gen_year, month), [])
                        if neg not in existing:
                            generated_dict[(gen_year, month)] = existing + [neg]
                        continue
                    first_week_day, last_day_no = calendar.monthrange(gen_year, month)
                    d1 = datetime.datetime(gen_year, month, 1)
                    d2 = datetime.datetime(gen_year, month, last_day_no)
                    if neg < 0 and neg >= -7:                                       # Every specified weekday
                        for i in self.day_in_range(d1, d2, abs(neg)):
                            existing = generated_dict.get((gen_year, month), [])
                            if i.day not in existing:
                                generated_dict[(gen_year, month)] = existing + [i.day]
                        continue
                    if neg == -99:                                                  # Last day of the month
                        first_week_day, last_day_no = calendar.monthrange(gen_year, month)
                        existing = generated_dict.get((gen_year, month), [])
                        if last_day_no not in existing:
                            generated_dict[(gen_year, month)] = existing + [last_day_no]
                        continue
                    if neg == -98:                                                  # Last weekday of the month
                        first_week_day, last_day_no = calendar.monthrange(gen_year, month)
                        ld = datetime.date(gen_year, month, last_day_no)
                        while ld.isoweekday() > 5:                                  # Last day of month is not a weekday
                            ld -= datetime.timedelta(days=1)                        # deduct days to get to Friday
                        existing = generated_dict.get((gen_year, month), [])
                        if ld.day not in existing:
                            generated_dict[(gen_year, month)] = existing + [ld.day]
                        continue
                    if neg <= -11 and neg >= -75:                        # Occurrence of a weekday
                        if neg in mondays:                               # Monday 1-5
                            map_idx = 0
                            occ = neg + abs(mondays[0])
                        elif neg in tuesdays:                            # Tuesday 1-5
                            map_idx = 1
                            occ = neg + abs(tuesdays[0])
                        elif neg in wednesdays:                          # Wednesday 1-5
                            map_idx = 2
                            occ = neg + abs(wednesdays[0])
                        elif neg in thursdays:                           # Thursday 1-5
                            map_idx = 3
                            occ = neg + abs(thursdays[0])
                        elif neg in fridays:                             # Friday 1-5
                            map_idx = 4
                            occ = neg + abs(fridays[0])
                        elif neg in saturdays:                           # Saturday 1-5
                            map_idx = 5
                            occ = neg + abs(saturdays[0])
                        elif neg in sundays:                             # Sunday 1-5
                            map_idx = 6
                            occ = neg + abs(sundays[0])
                        else:                                            # Undefined
                            continue
                        week_map = [index for (index, item) in enumerate(day_map) if item[map_idx]]
                        if abs(occ) >= len(week_map):
                            occ = len(week_map) - 1
                        week_idx = week_map[abs(occ)]
                        map_day = day_map[week_idx][map_idx]
                        existing = generated_dict.get((gen_year, month), [])
                        if map_day not in existing:
                            generated_dict[(gen_year, month)] = existing + [map_day]
        return generated_dict

    def GenerateNotes(self, date_dict):
        ''' Generated on start and when the year changes
            This routine generates a new dictionary of Notes from the one passed in and returns the generated dictionary.
            This because the original passed in dictionary may include date codes e.g. -99 for the last of a month
             or -1 all Mondays or -23 the 3rd Tuesday, which need to be calculated for the given month in the given year.
            An added complication is that the year may be set to zero, denoting all years, so if the calendar year is changed,
             this routine is run again, to ensure that the dates are relevant to the current year.
            Plus the month might be set to zero, indicating a note for all months
            Because some of the notes are calculated, a date may have muliple notes, so the notes are accumulated, to form
            a single note entry, separated by a + sign in tooltips etc
            If Public Holidays are included, these too are recalculated for the current year.
        '''
        generated_dict = {}
        for year, month, day in date_dict:
            gen_year = year
            if gen_year == 0:               # Zero entry = All years, so generate dates for the currently selected year
                d = self.calendar.GetDate()
                gen_year = d.year
            if month <= 0:                  # Zero entry = All months, generate dates for every month in the year
                month_start = 1
                month_end = 13
            else:
                month_start = month         # Valid month - restrict just to this month
                month_end = month + 1
            for month in range(month_start, month_end):
                day_map = calendar.monthcalendar(gen_year, month)
                note = date_dict.get((year, month, day))            # Get the note
                if not note:
                    note = date_dict.get((year, 0, day))            # if fail - get note for every month
                if day >= 0:
                    use_note = generated_dict.get((gen_year, month, day), '')
                    if use_note:
                        use_note = use_note+"\n+ "+note
                    else:
                        use_note = note
                    generated_dict[(gen_year, month, day)] = use_note
                    continue
                first_week_day, last_day_no = calendar.monthrange(gen_year, month)
                d1 = datetime.datetime(gen_year, month, 1)
                d2 = datetime.datetime(gen_year, month, last_day_no)
                if day < 0 and day >= -7:                                       # Every specified weekday
                    for i in self.day_in_range(d1, d2, abs(day)):
                        use_note = generated_dict.get((gen_year, month, i.day), '')
                        if use_note:
                            use_note = use_note+"\n+ "+note
                        else:
                            use_note = note
                        generated_dict[(gen_year, month, i.day)] = use_note
                    continue
                if day == -99:                                                  # Last day of the month
                    first_week_day, last_day_no = calendar.monthrange(gen_year, month)
                    use_note = generated_dict.get((gen_year, month, last_day_no), '')
                    if use_note:
                        use_note = use_note+"\n+ "+note
                    else:
                        use_note = note
                    generated_dict[(gen_year, month, last_day_no)] = use_note
                    continue
                if day == -98:                                                  # Last weekday of the month
                    first_week_day, last_day_no = calendar.monthrange(gen_year, month)
                    ld = datetime.date(gen_year, month, last_day_no)
                    while ld.isoweekday() > 5:                                  # Last day of month is not a weekday
                        ld -= datetime.timedelta(days=1)                        # deduct days to get to Friday
                    use_note = generated_dict.get((gen_year, month, ld.day), '')
                    if use_note:
                        use_note = use_note+"\n+ "+note
                    else:
                        use_note = note
                    generated_dict[(gen_year, month, ld.day)] = use_note
                    continue
                if day <= -11 and day >= -75:                            # Occurrence of a weekday
                    if day in mondays:                               # Monday 1-5
                        map_idx = 0
                        occ = day + abs(mondays[0])
                    elif day in tuesdays:                            # Tuesday 1-5
                        map_idx = 1
                        occ = day + abs(tuesdays[0])
                    elif day in wednesdays:                          # Wednesday 1-5
                        map_idx = 2
                        occ = day + abs(wednesdays[0])
                    elif day in thursdays:                           # Thursday 1-5
                        map_idx = 3
                        occ = day + abs(thursdays[0])
                    elif day in fridays:                             # Friday 1-5
                        map_idx = 4
                        occ = day + abs(fridays[0])
                    elif day in saturdays:                           # Saturday 1-5
                        map_idx = 5
                        occ = day + abs(saturdays[0])
                    elif day in sundays:                             # Sunday 1-5
                        map_idx = 6
                        occ = day + abs(sundays[0])
                    else:                                            # Undefined
                        continue
                    week_map = [index for (index, item) in enumerate(day_map) if item[map_idx]]
                    if abs(occ) >= len(week_map):
                        occ = len(week_map) - 1
                    week_idx = week_map[abs(occ)]
                    map_day = day_map[week_idx][map_idx]
                    use_note = generated_dict.get((gen_year, month, map_day), '')
                    if use_note:
                        use_note = use_note+"\n+ "+note
                    else:
                        use_note = note
                    generated_dict[(gen_year, month, map_day)] = use_note

        # If public holidays are available write them into the notes
        if holidays_available and self._cal_PublicHolidays:
            country, subdiv, language = self._cal_PublicHolidays
            d = self.calendar.GetDate()
            for k, v in holidays.country_holidays(country=country, subdiv=subdiv, years=d.year, language=language).items():
                use_note = generated_dict.get((k.year, k.month, k.day), '')
                if use_note:
                    use_note = use_note+"\n+ * "+v
                else:
                    use_note = " * "+v
                generated_dict[(k.year, k.month, k.day)] = use_note

            for item in self._cal_AddPublicHolidays:
                country, subdiv, language = item
                for k, v in holidays.country_holidays(country=country, subdiv=subdiv, years=d.year, language=language).items():
                    use_note = generated_dict.get((k.year, k.month, k.day), '')
                    if use_note:
                        use_note = use_note+"\n+ * "+v+' [ '+' '.join(item).title()+"]"
                    else:
                        use_note = " * "+v+' [ '+' '.join(item).title()+"]"
                    generated_dict[(k.year, k.month, k.day)] = use_note

        return generated_dict

    def OnToolTip(self, event):
        '''
        Test for date range restrictions.
        If there are Notes or Restricted entries for the day, generate and display tooltips
        '''
        self.calendar.SetToolTip('')
        if event.GetClassName() == 'wxMouseEvent':
            pos = event.GetPosition()
            pos_code, pos_date, pos_day = self.calendar.HitTest(pos)
            if pos_code != wx.adv.CAL_HITTEST_DAY and pos_code != wx.adv.CAL_HITTEST_HEADER:
                return
        else:
            pos_date = event.Date
            pos_code = wx.adv.CAL_HITTEST_NOWHERE

        # Position on calendar header report Holidays for which countries
        if pos_code == wx.adv.CAL_HITTEST_HEADER:
            lcl = self.locale.GetLocale()
            if holidays_available and self._cal_PublicHolidays:
                country, subdiv, language = self._cal_PublicHolidays
            else:
                self.calendar.SetToolTip("No Public Holidays\nLocale = "+lcl)
                return
            msg = ''
            msg = "Public Holidays for "+self.country_name
            if subdiv:
                msg += " region "+subdiv
            for item in self._cal_AddPublicHolidays:
                country, subdiv, language = item
                for c in holidays.registry.COUNTRIES.values():
                    if country in c:
                        name = c[0]
                        msg += "\n+ "+name
                        break
            self.calendar.SetToolTip(msg+"\nLocale = "+lcl)
            return

        range_check, lower, upper = self.calendar.GetDateRange()
        if range_check:
            if (lower != wx.DefaultDateTime and pos_date.IsEarlierThan(lower)) or \
                (upper != wx.DefaultDateTime and pos_date.IsLaterThan(upper)):
                msg = str(self._formatter(pos_date)).title()+'\n'+"Out of Range\n"
                if lower != wx.DefaultDateTime:
                    msg += str(lower.Format("%d-%b-%Y")).title()+' > '
                else:
                    msg += "Any date > "
                if upper != wx.DefaultDateTime:
                    msg += str(upper.Format("%d-%b-%Y")).title()
                else:
                    msg += "Any date"
                self.calendar.SetToolTip(msg)
                return

        restricted = self.RestrictedDates.get((pos_date.year, pos_date.month + 1), [])
        restricted_set = False
        if pos_date.day in restricted:
            restricted_set = True
        if self._cal_SetOnlyWeekDays and not pos_date.IsWorkDay():
            restricted_set = True
        d = (pos_date.year, pos_date.month + 1, pos_date.day)
        note = self.Notes.get(d, '')                        # Year/Month/Day specific Note or blank
        if restricted_set:
            note = "** Restricted **\n\n"+ note
        if not note:
            self.calendar.SetToolTip('')
        else:
            self.calendar.SetToolTip(str(self._formatter(pos_date)).title()+'\n'+note)

    def OnRightClick(self, event):
        '''
        If Right click display all Notes for the month
        '''
        click_date = self.calendar.GetDate()
        hol0 = self.Holidays.get((click_date.year,click_date.month + 1), [])            #Calendar holidays
        hol1 = self.Holidays.get((0,click_date.month + 1), [])                          #Fixed holidays
        hol2 = self.PublicHolidays.get((click_date.year,click_date.month + 1), [])      #Public holidays
        popmenu = wx.Menu()
        if holidays_available and self._cal_PublicHolidays:
            country, subdiv, language = self._cal_PublicHolidays
        else:
            country = subdiv = language = ''
        hdr = msg = ''
        if country:
            hdr = "Events for "+click_date.GetMonthName(click_date.month)+" "
            hdr += str(click_date.year)+"\nPublic Holidays for "+self.country_name
            if subdiv:
                hdr += " region "+subdiv
        else:
            hdr = 'Events for '+click_date.GetMonthName(click_date.month)+" "+str(click_date.year)

        for item in self._cal_AddPublicHolidays:
            country, subdiv, language = item
            for c in holidays.registry.COUNTRIES.values():
                if country in c:
                    name = c[0]
                    msg += "\n+ "+name
                    break
        popmenu.Append(wx.ID_ANY, hdr+msg)
        popmenu.AppendSeparator()

        for k, v in sorted(self.Notes.items()):
            if k[0] == click_date.year and k[1] == click_date.month + 1:
                if k[2] in hol0 or k[2] in hol1 or k[2] in hol2:
                    bmp = cal_fiesta.Bitmap
                else:
                    bmp = cal_event.Bitmap
                msg = str(k[2]).zfill(2)+ " "+ v
                m1 = wx.MenuItem(popmenu, wx.ID_ANY, msg)
                m1.SetBitmap(wx.Bitmap(bmp))
                popmenu.Append(m1)
        mlen = popmenu.GetMenuItemCount()
        if mlen <= 2:
            m1 = wx.MenuItem(popmenu, wx.ID_ANY, "No events or holidays")
            m1.SetBitmap(wx.Bitmap(cal_noevent.Bitmap))
            popmenu.Append(m1)
        self.PopupMenu(popmenu)
        popmenu.Destroy()


class YearPopup(wx.Frame):
    def __init__(self, parent, txt=''):
        wx.Frame.__init__(self, parent, wx.ID_ANY, title="Year Calendar", size=(600, 660))
        panel = wx.Panel(self)
        font = wx.Font(10, wx.FONTFAMILY_TELETYPE, wx.NORMAL, wx.NORMAL, False)
        tc = wx.TextCtrl(panel, -1, value=txt, size=(560, 640), style=wx.TE_MULTILINE|wx.TE_READONLY)
        tc.SetFont(font)
        butt = wx.Button(panel, -1, "&Quit")
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(tc, 0, wx.ALL|wx.EXPAND, 10)
        sizer.Add(butt, 0, wx.ALL, 10)
        panel.SetSizer(sizer)
        sizer.Fit(panel)
        sizer.Fit(self)
        self.Layout()
        butt.SetFocus()
        butt.Bind(wx.EVT_BUTTON, self.OnQuit)
        self.Bind(wx.EVT_CLOSE, self.OnQuit)
        self.Show()

    def OnQuit(self, event):
        self.Destroy()


class DemoFrame(wx.Frame):
    '''
        This demonstration code attempts to provide at least one example of every option, even if it's commented out
        It may offer examples of various options for the same thing, which explains its rather messy look
        The bulk of the marked dates, holidays, restrictions and notes are set around August 2023, when the testing
        was performed, so feel free to navigate to that month or change the values.
    '''
    def __init__(self, parent):
        wx.Frame.__init__(self, parent, -1, size=(400, 300), title="Calender Demo")

        #format = (lambda dt:
        #    (f'{dt.GetWeekDayName(dt.GetWeekDay())} {str(dt.day).zfill(2)}/{str(dt.month+1).zfill(2)}/{dt.year}')
        #    )

        #format = (lambda dt: (f'{dt.Format("%a %d-%m-%Y")}'))

        #Using a strftime format converting wx.DateTime to datetime.datetime
        #format = (lambda dt: (f'{wx.wxdate2pydate(dt).strftime("%A %d-%B-%Y")}'))

        format = (lambda dt: (f'{dt.Format("%A %d %B %Y")}'))

        panel = wx.Panel(self)

        self.ctl = CalendarCtrl(panel, -1, date=0, formatter=format, calendar_locale=wx.LANGUAGE_GERMAN)

        x=datetime.datetime.now()

        self.ctl.SetValue(0)

        font = wx.SystemSettings.GetFont(wx.SYS_SYSTEM_FONT)
        font.SetFractionalPointSize(16)
        self.ctl.SetCalendarFont(font)                                              # Set Calendar Font
        self.ctl.SetCalendarStyle(wx.adv.CAL_SHOW_WEEK_NUMBERS|wx.adv.CAL_MONDAY_FIRST|wx.adv.CAL_SHOW_HOLIDAYS)
        self.ctl.SetBackgroundColour('#e1ffe1')                                     # lightgreen
        self.ctl.SetCalendarHeaders(colFg='#ff0000', colBg='#90ee90')               # red/lightgreen
        self.ctl.SetWeekendColours(colFg='#ff0000', colBg='')                       # Weekends red/None
        self.ctl.SetCalendarHolidayColours(colFg=wx.NullColour, colBg='#ffff00')    # Holidays None/Yellow
        self.ctl.SetCalendarBg(colBg='#f0ffff')                                     # Background Colour Azure
        # Mark Border NONE, SQUARE or ROUND + Colour + foreground colour and background colour
        self.ctl.SetCalendarMarkBorder(border=wx.adv.CAL_BORDER_SQUARE, bcolour=wx.BLUE, colFg=wx.NullColour, colBg=wx.NullColour)
        # Notes Border NONE, SQUARE or ROUND + Colour + foreground colour and background colour
        self.ctl.SetCalendarMarkNotes(border=wx.adv.CAL_BORDER_SQUARE, bcolour=wx.GREEN, colFg=wx.NullColour, colBg='#ffbf00')

        # Some other options
        #self.ctl.SetFormatter(format)                                              # Set format separately
        #self.ctl.SetValue(x.timestamp())                                           # Set Date
        #self.ctl.SetValue(wx.DateTime.Today())
        #self.ctl.SetCalendarHighlights(colFg='#ff0000', colBg='#90ee90')           # Highlight selected day red/lightgreen
        #self.ctl.SetCalendarFg(colFg='#0000ff')                                    # Foreground Colour Blue
        #self.ctl.SetCalendarMarkNotes(border=0, bcolour=wx.NullColour)             # Turn off borders for notes
        #self.ctl.SetCalendarOnlyWeekDays(True)                                     # Only non holiday weekdays are selectable
        #self.ctl.SetToolTip('Struck through dates are not selectable')

        self.ctl.SetCalendarMarkDates({
                                       (0, 1) : [-11,-23,1],                   # 1st Monday, 3rd Tuesday and the 1st January
                                       (x.year, 7) : [2,5,30],
                                       (x.year, 8) : [12,20,27],
                                       (0, 9) : [1,27,-98]                     # 1st, 27th and the last weekday of September
                                      })

        self.ctl.SetCalendarHolidays({
                                     (0, 1) : [1,],                                 # January 1st every year
                                     (0, 2) : [1],
                                     (x.year, 8) : [15],
                                     (0, 12) : [25,26]                              # 25th & 26th December every year
                                    })

        self.ctl.SetCalendarNotes({
                                     #(0, 0, -55) : "Last Friday of the month - Pay Day!",# Last Friday, every month. every year
                                     (0, 1, 1) : "New Year's Day\n2nd Note",             # January 1st every year
                                     (0, 1, -11) : "First Monday of the year",
                                     (0, 1, -23) : "3rd Tuesday of the year",
                                     (0, 1, -99) : "Last day of January",
                                     (0, 1, -35) : "The last Wednesday of January",
                                     (0, 2, 1) : "1st February",
                                     (x.year, 2, -5)  : "Every Friday in February",      # This year only
                                     (x.year, 8, 7) : "Marked for no reason whatsoever", # This year only
                                     (x.year, 8, 15) : "A holiday August 15 2023",       # This year only
                                     (x.year, 8, 20) : "Marked for reason X",            # this year only
                                     (0, 9, -98) : "Last weekday of September",
                                     (0, 12, 25) : "Merry Christmas!",
                                     (0, 2, -99) : "Last day of February"
                                    })

        self.ctl.SetCalendarRestrictDates({
                                       (0, 1) : [-1, -2, 5],     # exclude Mondays and Tuesdays, 5th of January All years
                                       (0, 2) : [-99,],          # exclude last day of February - All years
                                       (0, 8) : [-98,]           # exclude last weekday of August - All years
                                        })

        # Restrict Calendar to a date range (define a lowerdate or an upperdate or both
        #self.ctl.SetCalendarDateRange(lowerdate=wx.DateTime(23,8,2023), upperdate=wx.DateTime(23,9,2023))

        # Public Holidays requires the python holidays module (pip install --upgrade holidays)
        self.ctl.AddPublicHolidays(country="DE", subdiv="ST", language=())    # Germany region Lower Saxony
        self.ctl.AddPublicHolidays(country="GB", subdiv="ENG", language="")   # Great Britain region England
        self.ctl.AddPublicHolidays(country="ES", subdiv="AN", language="")    # Spain region Andalucia
        #self.ctl.AddPublicHolidays(country="JP", subdiv="", language="en_US")    # Japan

        self.ctl.Bind(wx.adv.EVT_CALENDAR_MONTH, self.OnCalEvt)
        self.ctl.Bind(wx.adv.EVT_CALENDAR_YEAR, self.OnCalEvt)
        self.ctl.Bind(wx.adv.EVT_CALENDAR, self.OnCalEvt)
        self.ctl.Bind(wx.adv.EVT_CALENDAR_SEL_CHANGED, self.OnCalEvt)
        self.ctl.Bind(EVT_DATE_CHANGED, self.OnEvent)
        sizer=wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(self.ctl, 0, wx.EXPAND, 0)
        panel.SetSizerAndFit(sizer)
        self.Centre()

    def OnEvent(self, event):
        # ExtendedCalendarCtrl events
        print("Extended Calendar Event", event.GetEventType())
        obj = event.GetEventObject()
        print("\nevt", event.GetValue())

        x = event.GetDate() # wx.DateTime object
        print("evt", x)
        print("Day", x.GetDayOfYear(),"Week", x.GetWeekOfYear(), "Name", x.GetWeekDayName(x.GetWeekDay()))

        print("Datetime", event.GetDateTime()) # datetime.datetime object
        print("Timestamp", event.GetTimeStamp())
        print("Value", event.GetValue())
        print("Notes", event.GetNote())
        print("Marked", event.IsMarked())
        print("Calendar Holiday", event.IsHoliday())
        print("Public Holiday", event.IsPublicHoliday())
        print("Locale", obj.GetLocale())

    def OnCalEvt(self, event):
        # Internal wx.CalendarCtrl events
        obj = event.GetEventObject()
        print("wxCalendar Event", event.GetEventType())

if __name__ == '__main__':
    app = wx.App()
    frame = DemoFrame(None)
    frame.Show()
    app.MainLoop()
